from __future__ import annotations

from typing import TYPE_CHECKING, Any, Final

from ..overloads import SINGLETON_OVERLOAD
from .record import FnImplement

if TYPE_CHECKING:
    from ..typing import RecordsT
    from .base import Fn


class FnCompose:
    singleton: Final = SINGLETON_OVERLOAD
    fn: Fn

    def __init__(self, fn: Fn):
        type(self).fn = fn

    def call(self, records: RecordsT, *args, **kwargs) -> Any:
        raise NotImplementedError

    def signature(self):
        return FnImplement(self.fn)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Generic, Protocol, TypeVar

from typing_extensions import Concatenate

from ..globals import iter_layout
from ..typing import CR, P, R
from .record import FnImplement

if TYPE_CHECKING:
    from .compose import FnCompose

FC = TypeVar("FC", bound="FnCompose", covariant=True)


class ExtractCall(Protocol[CR]):
    @property
    def call(self) -> CR: ...


class Fn(Generic[FC]):
    def __init__(self, compose: type[FC]):
        self.compose = compose(self)

    # def entity(self, impl: C) -> FnImplementEntity[C]:
    #     return FnImplementEntity(impl)

    @property
    def _(self):
        return type(self.compose)

    def __call__(self: Fn[ExtractCall[Callable[Concatenate[Any, P], R]]], *args: P.args, **kwargs: P.kwargs) -> R:  # type: ignore
        sign = FnImplement(self)

        for layout in iter_layout(self):
            if sign in layout.fn_implements:
                records = layout.fn_implements[sign]

            return self.compose.call(records, *args, **kwargs)
        else:
            raise NotImplementedError("cannot lookup any implementation with given arguments")



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Generator, Generic, Protocol, overload

from typing_extensions import Concatenate, Self

from ..typing import CQ, P1, P2, C, Call, CnQ, CnR, P, R, RecordsT, T
from .compose import FnCompose
from .harvest import FnHarvestControl
from .implement import FnImplementEntity
from .record import FnOverloadSignal

CollectEndpointTarget = Generator[FnOverloadSignal, None, T]


class EndpointCollectReceiver(Protocol[CnR]):
    @overload
    def __call__(self: EndpointCollectReceiver[None], value: C) -> FnImplementEntity[C]: ...
    @overload
    def __call__(self: EndpointCollectReceiver[C], value: FnImplementEntity[C]) -> FnImplementEntity[C]: ...
    @overload
    def __call__(self: EndpointCollectReceiver[C], value: C) -> FnImplementEntity[C]: ...


@dataclass(init=False, eq=True, unsafe_hash=True)
class FnCollectEndpoint(Generic[P, CnQ]):
    @overload
    def __init__(
        self: FnCollectEndpoint[P1, Callable[P2, R]],
        target: Call[Concatenate[Any, P1], CollectEndpointTarget[Callable[P2, R]]],
    ) -> None: ...

    @overload
    def __init__(
        self: FnCollectEndpoint[P1, None],
        target: Callable[Concatenate[Any, P1], CollectEndpointTarget[T]],
    ): ...

    def __init__(self, target):
        if TYPE_CHECKING:
            self.target = target
        else:
            self.target = target.__func__

    def __set_name__(self, owner: type[FnCompose], name: str):
        self.compose = owner

    @property
    def fn(self):
        return self.compose.fn

    @overload
    def __get__(self: FnCollectEndpoint[P1, None], instance: None, owner: type) -> Callable[P1, EndpointCollectReceiver[Callable]]: ...

    @overload
    def __get__(self: FnCollectEndpoint[P1, CQ], instance: None, owner: type) -> Callable[P1, EndpointCollectReceiver[CQ]]: ...

    @overload
    def __get__(self, instance: Any, owner) -> Self: ...

    def __get__(self, instance, owner) -> Any:
        if instance is None:
            return self.get_collect_entity

        return self

    @overload
    def get_control(self: FnCollectEndpoint[..., C], records: RecordsT) -> FnHarvestControl[C]: ...
    @overload
    def get_control(self: FnCollectEndpoint[..., None], records: RecordsT) -> FnHarvestControl[Callable]: ...

    def get_control(self, records: RecordsT) -> FnHarvestControl:
        return FnHarvestControl(self, records)

    def get_collect_entity(self, *args: P.args, **kwargs: P.kwargs):
        def receiver(entity: C | FnImplementEntity[C]) -> FnImplementEntity[C]:
            if not isinstance(entity, FnImplementEntity):
                entity = FnImplementEntity(entity)

            entity.add_target(self, self.compose, *args, **kwargs)
            return entity

        return receiver



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generic

from ..typing import C, RecordsT
from .overload import FnOverload, TCallValue

if TYPE_CHECKING:
    from .endpoint import FnCollectEndpoint


class FnHarvestControl(Generic[C]):
    def __init__(self, endpoint: FnCollectEndpoint, records: RecordsT) -> None:
        self.endpoint = endpoint
        self.records = records

    @property
    def record(self):
        if self.endpoint not in self.records:
            raise NotImplementedError(f"endpoint {self.endpoint} not found in records")

        return self.records[self.endpoint]

    def use(self: FnHarvestControl[C], overload: FnOverload[Any, Any, TCallValue], value: TCallValue):
        return FnHarvest(self).apply(overload, value)


class FnHarvest(Generic[C]):
    def __init__(self, control: FnHarvestControl[C]) -> None:
        self.control = control
        self.result = None

    def apply(self, overload: FnOverload[Any, Any, TCallValue], value: TCallValue):
        self.result = overload.dig(self.control.record, value)
        return self

    def use(self, overload: FnOverload[Any, Any, TCallValue], value: TCallValue):
        if self.result is None:
            raise NotImplementedError(f"result is None, cannot use overload {overload} with value {value}")

        self.result = dict(self.result.items() & overload.dig(self.control.record, value).items())
        return self

    @property
    def first(self) -> C:
        if self.result is None:
            raise NotImplementedError("result is None, cannot get first item")

        return next(iter(self.result))  # type: ignore

    def __iter__(self) -> list[C]:
        if self.result is None:
            raise NotImplementedError("result is None, cannot iterate")

        return list(self.result)  # type: ignore

    def __bool__(self):
        return bool(self.result)



from .context import CollectContext as CollectContext
from .context import InstanceContext as InstanceContext
from .fn import Fn as Fn
from .fn import FnCompose as FnCompose
from .fn import FnImplementEntity as FnImplementEntity
from .fn import FnOverload as FnOverload
from .fn import FnRecord as FnRecord
from .globals import global_collect as global_collect
from .globals import local_collect as local_collect
from .instance_of import InstanceOf as InstanceOf
from .overloads import SimpleOverload as SimpleOverload
from .overloads import TypeOverload as TypeOverload
from .scoped import scoped_collect as scoped_collect
